import json
import traceback

from greenstalk import Client, Job, TimedOutError
from gwerks import emitter, uid

from fwq.broker import parse_broker_str, Broker
from fwq.logs import JobLogger, Sender
from fwq.constants import JOB_TYPE_WORKER_CONTROL, WORKER_CONTROL_ACTION_RETIRE


@emitter()
class Worker:
    def __init__(self, for_app, broker: str = Broker.DEFAULT, worker_id=None):
        if not worker_id:
            worker_id = uid()
        self._uid = worker_id
        self._for_app = for_app
        self._broker = broker

        broker_tpl = parse_broker_str(broker)
        self._client = Client(broker_tpl, use=for_app, watch=[for_app])
        print(f"Worker [{self._uid}] is connected to {broker} doing work for [{for_app}]")

        self._work_loop_running = False
        self._job_type_map = {
            JOB_TYPE_WORKER_CONTROL: "self._handle_worker_control"
        }

        self._job_logger = JobLogger(self._broker, self._uid)

    def register_log_sender(self, log_sender: Sender):
        self._job_logger.register_logstream(log_sender)

    def do_jobs(self, wait_for_job_secs=60):
        self._work_loop_running = True
        while self._work_loop_running:
            try:
                self.do_job(wait_for_job_secs)

            except Exception as e:
                traceback.print_exc()
                print(f"Unexpected ERROR in work loop: {e}")

    def do_job(self, wait_for_job_secs=60):

        # the_tube = make_tube_name(self._for_app, self._use_tube_name)
        the_tube = self._for_app
        print(f"Worker [{self._uid}] waiting for job from Broker [{self._broker}.{the_tube}]")

        job_body, job_id = self._reserve_job(timeout=wait_for_job_secs)

        # no job, timeout expired
        if job_id is None:
            return

        if job_body is None:
            raise Exception(f"job_body is empty for job {job_id}")

        if "type" not in job_body:
            raise Exception(f"'type' not in job_body for job {job_id}")
        job_type = job_body['type']

        if "name" not in job_body:
            raise Exception(f"'name' not in job_body for job {job_id}")
        job_name = job_body['name']

        job_data = {}
        if 'data' in job_body:
            job_data = job_body['data']

        # with JobPrint(self._beanstalk_server_id, self._uid, job_id):
        self._job_logger.change_jobs(job_id)
        with self._job_logger:
            try:

                print(f"---------------------------------------------------------------------------------")
                print(f"{job_name} is STARTING")

                if job_type in self._job_type_map.keys():
                    job_type = self._job_type_map[job_type]
                print(f"job_type: {job_type}")
                print(f"job_data: {job_data}")

                code = ""
                try:
                    job_type = str(job_type)
                    ri = job_type.rindex(".")
                    code += f"from {job_type[:ri]} import {job_type[ri + 1:]}; " \
                            f"globals()['result'] = {job_type[ri + 1:]}(job_info)"
                except ValueError:
                    code += f"globals()['result'] = {job_type}(job_info)"

                print(f"exec: {code}")

                print(f"---------------------------------------------------------------------------------")
                print(f"{job_name} is RUNNING")

                g = {"job_info": {"job_data": job_data, "job_id": job_id}}
                exec(code, g)
                result = g['result']

                print(f"---------------------------------------------------------------------------------")
                print(f"{job_name} is EXITING")
                print(f"result: {result}")

                if result is not None and result is False:
                    self._client.release(Job(job_id, job_body))
                    print(f"resolution: not handled and released")

                else:
                    self._client.delete(job_id)
                    print(f"resolution: succeeded and deleted")

            except Exception as e:
                result = {"error": str(e)}
                print(f"---------------------------------------------------------------------------------")
                print(f"{job_name} is EXITING")
                print(f"result: {result}")
                try:
                    self._client.bury(Job(job_id, job_body))
                    print(f"resolution: failed and buried")
                except Exception as ex:
                    print(f"ERROR when burying job ({job_id}) {job_name}")

            finally:
                print(f"---------------------------------------------------------------------------------")

    def _reserve_job(self, timeout=None):
        try:
            gs_job = self._client.reserve(timeout)
            job_id = gs_job.id
            job_body = json.loads(gs_job.body)
            if "type" not in job_body.keys():
                raise Exception(f"'type' not found in the job body for {gs_job} from {self._for_app}")
            # if "data" not in job_body.keys():
            #     raise Exception(f"'data' not found in the job body for {gs_job} from {self._for_app}")
            # if "name" not in job_body.keys():
            #     raise Exception(f"'name' not found in the job body for {gs_job} from {self._for_app}")
            # job_type = job_body['type']
            # job_data = job_body['data']
            return job_body, job_id
        except TimedOutError:
            return None, None

    def _handle_worker_control(self, job_data):
        worker_id = job_data['worker_id']
        if worker_id != self._uid:
            print(f"Specified worker_id ({worker_id}) does not match this worker ({self._uid}), "
                  f"job is for a different worker")
            return False
        action = job_data['action']
        if action == WORKER_CONTROL_ACTION_RETIRE:
            print(f"Worker [{self._uid}] has been directed to retire")
            self._work_loop_running = False

        else:
            raise Exception(f"Unsupported worker control action: {action}")

    # def _client(self) -> Client:
    #     if self._bs_client is None:
    #         self._bs_client, self._broker = get_beanstalk_client(self._beanstalk_host, self._for_app)
    #     return self._bs_client



import sys
import yaml
import json

from gwerks.decorators import emitter
from gwerks.cli import cli, Clo
# from fwq.the_broker import Broker
# from fwq.worker import Worker

from fwq import *
from fwq.api import start, stop, do_next_job, do_jobs, kick_job, stats, nq, purge, peek

"""
context = DockerContext(data={
    "dockerfile_str": Broker.DOCKERFILE_BASE + "COPY gwerks.tar.gz .\nCOPY fwq.tar.gz .\nRUN pip3 install gwerks.tar.gz\nRUN pip3 install fwq.tar.gz\n",
    "files": [("../gwerks/dist/gwerks-25.2.13.tar.gz", "gwerks.tar.gz"), ("dist/fwq-0.0.0.tar.gz", "fwq.tar.gz")]
})

broker = Broker(context=context, map_to_host_port=11300)
"""


# --------------------------------------------------------------------------- #
# CLI entry point
# --------------------------------------------------------------------------- #
@emitter()
def fwq():

    _debug_traceback_limit = 1000
    sys.tracebacklimit = 0

    try:

        # command line arguments and default values
        clo = cli([
            {   # base args
                "action": Clo.REQUIRED,
                "debug": None
            },
            {   # action start, stop
                "name": Broker.DEFAULT_NAME
            },
            {   # action start
                "port_spec": Broker.DEFAULT_PORT,
                "host_volume": "./data",
                "max_message_size": "65535",
                "log_group_prefix": None,
                "net_name": None,
                "net_driver": DockerNetwork.DEFAULT_DRIVER,
                "context": None
            },
            {  # action do_next_job, kick_job, stats, nq, purge, peek
                "for": Clo.REQUIRED,
                "broker": f"{Broker.DEFAULT_NAME}:{Broker.DEFAULT_PORT}",
            },
            {  # action do_next_job
                "wait": "60"
            },
            {  # action kick_job, stats, peek
                "id": None
            },
            {  # action nq
                "job_type": Clo.REQUIRED,
                "job_data": None,
                "job_name": None,
                "priority": "65536",
                "delay_secs": "0",
                "ttr": "60",
            },
            {  # action kick_job, purge, peek
                "at_most": Clo.REQUIRED,
                "in_state": JOB_STATE_READY,
            }
        ])
        print(f"Hello!")

        debug = clo.get("debug") == "True"
        if debug:
            sys.tracebacklimit = _debug_traceback_limit

        action = clo.get("action")
        cli_action = f"cli_{action}"

        print(f"---------------------------------------------------------------------------------")

        if not action:
            raise Exception("--action | -a is required")

        globals()[cli_action](clo)

        print(f"---------------------------------------------------------------------------------")

        # sys.exit(0)

    except Exception as e:
        msg = f"ERROR: {e}"
        print(msg)
        # sys.exit(1)

    finally:
        pass


@emitter()
def cli_start(clo: Clo):
    broker_name = clo.get("name")
    broker_port = clo.get("port_spec")
    host_volume = clo.get("host_volume")
    max_message_size = clo.get("max_message_size")
    log_group_prefix = clo.get("log_group_prefix")
    net_name = clo.get("net_name")
    net_driver = clo.get("net_driver")
    context_yaml_file = clo.get("context")
    start(broker_name, broker_port, net_name, net_driver, host_volume, max_message_size,
          log_group_prefix, context_yaml_file)


@emitter()
def cli_stop(clo: Clo):
    broker_name = clo.get("name")
    stop(broker_name)


@emitter()
def cli_do_next_job(clo: Clo):
    for_app = clo.get("for")
    broker_host = clo.get("broker")
    wait_secs = int(clo.get("wait"))
    do_next_job(for_app, broker_host, wait_secs)


@emitter()
def cli_do_jobs(clo: Clo):
    for_app = clo.get("for")
    broker_host = clo.get("broker")
    wait_secs = int(clo.get("wait"))
    do_jobs(for_app, broker_host, wait_secs)


@emitter()
def cli_kick_job(clo: Clo):
    for_app = clo.get("for")
    broker_host = clo.get("broker")
    job_id = clo.get("id")
    kick_count = clo.get("at_most")
    kick_job(for_app, broker_host, job_id=job_id, kick_count=kick_count)


@emitter()
def cli_stats(clo: Clo):
    for_app = clo.get("for")
    broker_host = clo.get("broker")
    job_id = clo.get("id")
    return stats(for_app, broker_host, job_id=job_id)


@emitter()
def cli_nq(clo: Clo):
    for_app = clo.get("for")
    broker_host = clo.get("broker")
    job_type = clo.get("job_type")
    job_data = clo.get("job_data")
    job_name = clo.get("job_name")
    priority = int(clo.get("priority"))
    delay_secs = int(clo.get("delay_secs"))
    ttr = int(clo.get("ttr"))
    nq(for_app, job_type, broker_host, job_data, job_name, priority, delay_secs, ttr)


@emitter()
def cli_purge(clo: Clo):
    for_app = clo.get("for")
    the_broker = clo.get("broker")
    at_most = clo.get("at_most")
    in_state = clo.get("in_state")
    purge(for_app, the_broker, at_most, in_state)


@emitter()
def cli_peek(clo: Clo):
    for_app = clo.get("for")
    the_broker = clo.get("broker")
    job_id = clo.get("id")
    next_in_state = clo.get("in_state")
    job = peek(for_app, the_broker, job_id, next_in_state)
    print(yaml.dump(json.loads(job.body), allow_unicode=True, default_flow_style=False))
